class Output:
    """
    Implement this to support formating the collected data in some kind of output format.
    Example output formats could be stdout printouts, a .tsv file, or a .png file with a graph.
    """

    def __init__(self, args):
        self.args = args

    # Called in the beginning of processing,
    # to announce what args to use for processing.
    def start(self, columns):
        pass

    # Called each time a commit has been analyzed,
    # and it is time to present the data in the commit.
    def add_row(self, date, column_to_lines_dict):
        pass

    # Called when the processing of commits is complete.
    # This is a good time to write output files to disk.
    def finish(self):
        pass



#!/usr/bin/env python3

# pylint: disable=C0116

import sys

from datetime import datetime
import os
import os.path
import pygit2

from .args import get_args
from .matplotlib_output import MatplotlibOutput
from .progress import Progress
from .separated_values_output import SeparatedValuesOutput
from .utils import (
    get_extensions_sorted_by_popularity,
    get_top_three_extensions,
    to_relative_numbers_if_enabled,
)


def main():
    args = get_args()
    if args.list:
        list_available_file_extensions(args)
    else:
        outputs = get_outputs(args)
        process_commits(args, outputs)


def list_available_file_extensions(args):
    ext_to_lines = get_data_for_first_commit(args)
    sorted_exts = get_extensions_sorted_by_popularity(ext_to_lines)
    print("Available extensions in first commit:")

    len_of_longest_ext = len(max(sorted_exts, key=len))
    for ext in sorted_exts:
        print(f"{ext:<{len_of_longest_ext}} - {ext_to_lines[ext]} lines")


def get_outputs(args):
    # TODO: Support multiple --output arguments
    outputs = []

    # TODO: CLI tests
    if args.output_ext == ".svg" or args.output_ext == ".png":
        outputs.append(MatplotlibOutput(args))
    elif args.output_ext == ".tsv":
        outputs.append(SeparatedValuesOutput(args, "\t"))
    elif args.output_ext == ".csv":
        outputs.append(SeparatedValuesOutput(args, ","))
    else:
        sys.exit(f"Output file format '{args.output_ext}' not supported")

    return outputs


def process_commits(args, outputs):
    columns = args.columns
    if len(columns) == 0:
        print("No file extensions specified, will use top three.")
        data = get_data_for_first_commit(args)
        columns = get_top_three_extensions(data)
        print(f"Top three extensions were: {' '.join(columns)}")

    if len(columns) == 0:
        sys.exit("No extensions to count lines for")

    ext_to_column = generate_ext_to_column_dict(columns)

    commits_to_process = get_commits_to_process(args)
    progress_state = Progress(args, len(commits_to_process))

    # Print column headers
    for output in outputs:
        output.start(columns)

    # Print rows
    for commit in commits_to_process:
        date = get_commit_date(commit)
        column_to_lines_dict = process_commit(commit, ext_to_column, progress_state)

        for output in outputs:
            output.add_row(
                columns,
                date,
                to_relative_numbers_if_enabled(args, column_to_lines_dict),
            )

        progress_state.commit_processed()

    # Wrap things up
    for output in outputs:
        output.finish()


# Calls process_commit for the first commit (possibly from --first-commit)
def get_data_for_first_commit(args):
    repo = get_repo()
    rev = repo.revparse_single(args.first_commit)
    return process_commit(rev.peel(pygit2.Commit), None, Progress(args, 1))


def get_commits_to_process(args):
    commits_to_process = []

    rows_left = args.max_commits

    date_of_last_row = None
    try:
        for commit in get_git_log_walker(args):
            if rows_left == 0:
                break

            # Make sure --min-interval days has passed since last printed commit before
            # processing and printing the data for another commit
            current_date = commit.commit_time
            if enough_days_passed(args, date_of_last_row, current_date):
                date_of_last_row = current_date

                commits_to_process.append(commit)

                rows_left -= 1
    except KeyError:
        # Analyzing a shallow git clone will cause the walker to throw an
        # exception in the end. That is not a catastrophe. We already collected
        # some data. So just keep going after printing a notice.
        print("WARNING: unexpected end of git log, maybe a shallow git repo?")
        pass

    # git log shows most recent first, but in the graph
    # you want to have from oldest to newest, so reverse
    commits_to_process.reverse()

    return commits_to_process


def process_commit(commit, ext_to_column, progress_state):
    """
    Counts lines for files with the given file extensions in a given commit.
    """

    blobs = get_blobs_in_commit(commit)

    # Loop through all blobs in the commit tree
    column_to_lines = {}
    for index, (blob, ext) in enumerate(blobs, start=1):
        progress_state.print_state(index, len(blobs))

        # Figure out if we should count the lines for the file extension this
        # blob has, by figuring out what column the lines should be added to,
        # if any
        column = ext_to_column.get(ext) if ext_to_column else ext
        # If no specific columns are requested, we are probably invoked
        # with --list, so count the lines for all extensions

        # If the blob has an extension we care about, count the lines!
        if column:
            lines = get_lines_in_blob(blob)
            column_to_lines[column] = column_to_lines.get(column, 0) + lines

    return column_to_lines


def get_all_blobs_in_tree(tree):
    blobs = []
    trees_left = [tree]
    # Say no to recursion
    while len(trees_left) > 0:
        tree = trees_left.pop()
        for obj in tree:
            if isinstance(obj, pygit2.Tree):
                trees_left.append(obj)
            elif isinstance(obj, pygit2.Blob):
                blobs.append(obj)
    return blobs


def get_blobs_in_commit(commit):
    blobs = []
    for obj in get_all_blobs_in_tree(commit.tree):
        ext = os.path.splitext(obj.name)[1]
        if ext:
            blobs.append((obj, ext))

    return blobs


c = {}  # TODO: Make optinal and measure memory usage for linux kernel


def get_lines_in_blob(blob):
    if blob.oid in c:
        return c[blob.oid]

    lines = 0
    for byte in memoryview(blob):
        if byte == 10:  # \n
            lines += 1

    c[blob.oid] = lines
    return lines


def get_repo():
    return pygit2.Repository(os.environ.get('GIT_DIR', '.'))


def get_git_log_walker(args):
    repo = get_repo()

    rev = repo.revparse_single(args.first_commit)

    walker = repo.walk(rev.peel(pygit2.Commit).oid)

    if not args.all_parents:
        walker.simplify_first_parent()

    return walker


def enough_days_passed(args, date_of_last_row, current_date):
    """
    Checks if enough days according to --min-interval has passed, i.e. if it is
    time to process and print another commit.
    """

    if date_of_last_row:
        days = ((date_of_last_row - current_date) / 60 / 60 / 24)
        return days > args.min_interval_days
    return True


def generate_ext_to_column_dict(columns):
    extension_to_column_dict = {}
    for column in columns:
        for ext in column.split('+'):
            extension_to_column_dict[ext] = column
    return extension_to_column_dict


def get_commit_date(commit):
    return datetime.utcfromtimestamp(commit.commit_time).strftime('%Y-%m-%d')



from .output import Output
from .utils import print_file_written

"""
This module adds support for

  Tab-separated values

and

  Comma-separated values

output, i.e. .tsv and .csv file formats.
"""


def write_header_row(writer, separator, columns):
    # To get correct tab alignment, pad with spaces in place of a date
    # on the first row
    writer.write("          ")  # for YYYY-MM-DD

    # Now write the columns
    for column in columns:
        writer.write(f"{separator}{column}")

    # ... and finish with a newline
    writer.write("\n")


def write_row(writer, separator, columns, date, column_to_lines_dict):
    # Date
    writer.write(date)

    # Line count information
    for column in columns:
        writer.write(f"{separator}{column_to_lines_dict.get(column, 0)}")

    # ... and finish with a newline
    writer.write("\n")


class SeparatedValuesOutput(Output):

    def __init__(self, args, separator):
        super().__init__(args)
        self.separator = separator
        self.dest = open(args.output, 'w')

    def start(self, columns):
        write_header_row(self.dest, self.separator, columns)

    def add_row(self, columns, date, column_to_lines_dict):
        write_row(self.dest, self.separator, columns, date, column_to_lines_dict)

    def finish(self):
        print_file_written(self.args.output)
        self.dest.close()



import os
import sys
import argparse
import git_repo_language_trends
from pathlib import Path

desc = """
Description:
============

Analyze programming language usage over time in a git repository and produce a
graphical or textual representation of the result.

Available output file formats:
* .svg - Scalable Vector Graphics
* .png - Portable Network Graphics
* .csv - Comma-separated values
* .tsv - Tab-separated values

Examples:
=========

First go to any git repository:

    cd ~/src/any-git-repository

Run the tool without arguments to analyze programing language usage of top three
file extensions, and write the result to an SVG file:

    git-repo-language-trends

Analyze Objective-C vs Swift and write the result to a .csv file e.g. so you can
create a graph yourself in your spreadsheet software of choice:

    git-repo-language-trends .m+.h .swift --output=output.csv

Analyze Java vs Kotlin and write the result to a PNG file with a white
background and a custom size:

    git-repo-language-trends .java .kt --output=output.png --size-inches=10:6 --style=light

Arguments:
==========
"""


def positive_int(arg):
    i = int(arg)
    if i < 0:
        raise argparse.ArgumentTypeError("Must not be negative")
    return i


def positive_float(arg):
    i = float(arg)
    if i < 0:
        raise argparse.ArgumentTypeError("Must not be negative")
    return i


def formatter(prog):
    return argparse.RawDescriptionHelpFormatter(
        "git-repo-language-trends",
        indent_increment=4,
        max_help_position=38,
    )


def get_args():
    parser = argparse.ArgumentParser(
        description=desc,
        formatter_class=formatter,
    )

    parser.add_argument(
        "columns",
        metavar=".ext .ext+.ext",
        nargs='*',
        help="""For what file extensions lines will be counted. Can be specified
        multiple times. Use '.ext' for regular line counting. Use '.ext1+.ext2'
        syntax for auto-summation of several file extensions, e.g. .c+.h for all C files.
        If you specify no file extensions, the top three extensions in the
        repository will be used, based on the number of lines in files with the
        extensions.""",
    )

    parser.add_argument(
        '--version',
        action='version',
        version=f"%(prog)s {git_repo_language_trends.__version__}",
    )

    parser.add_argument(
        "--list", "-l",
        action='store_true',
        help="list file extensions and their total line count in the first commit",
    )

    parser.add_argument(
        "--min-interval-days",
        metavar="<int>",
        type=positive_int,
        default=7,
        help="""mimimum interval in days between analyzed commits
        (default: %(default)s)""",
    )

    parser.add_argument(
        "--max-commits", "-n",
        metavar="<int>",
        type=positive_int,
        default=sys.maxsize,
        help="""maximum number of commits to analyze
        (default: %(default)s)"""
    )

    parser.add_argument(
        "--first-commit",
        metavar="<rev>",
        default="HEAD",
        help="""the commit or tag or branch to start from
        (default: %(default)s)"""
    )

    parser.add_argument(
        "--relative",
        action='store_true',
        help="use relative instead of absolute numbers",
    )

    parser.add_argument(
        "--output", "-o",
        metavar="<out.ext>",
        default=get_default_output(),
        help="""output filename and format (via extension .svg .png .csv or .tsv)
        (default: %(default)s)""",
    )

    svg_group = parser.add_argument_group(
        "SVG/PNG related optional arguments",
    )

    svg_group.add_argument(
        "--size-inches",
        metavar="<w:h>",
        default="11.75:8.25",
        help="""width:height in inches of the diagram
        (default: %(default)s)""",
    )

    svg_group.add_argument(
        "--style",
        metavar="<name>",
        default="dark",
        choices=['dark', 'light'],
        help="""pass 'dark' for black background and 'light' for white background
        (default: %(default)s)""",
    )

    advanced_group = parser.add_argument_group(
        "advanced optional arguments",
    )

    advanced_group.add_argument(
        "--no-progress",
        action='store_true',
        help="""[ADVANCED] do not print progress""",
    )

    advanced_group.add_argument(
        "--all-parents", "-a",
        action='store_true',
        help="""[ADVANCED] increase pool of candidate commits by following all
        commit parents, but with the risk of producing inconsistent/jumpy graphs""",
    )

    args = parser.parse_args()

    # Figure out output file extension
    # Without an extension, we treat the entire filename as the extension
    name, ext = os.path.splitext(args.output)
    if not ext:
        ext = name
        name = ""
    args.output_ext = ext

    # Do a favor to the user, create the parent dirs if they are missing
    Path(os.path.dirname(args.output)).mkdir(exist_ok=True, parents=True)

    return args


def get_default_output():
    cwd = os.getcwd()
    basename = os.path.basename(cwd)

    return f"{basename}-language-trends.png"



from .output import Output


class MatplotlibOutput(Output):
    def __init__(self, args):
        super().__init__(args)
        self.columns = None
        self.rows = []

    # Called in the beginning of processing,
    # to announce what columns to use in the output.
    def start(self, columns):
        self.columns = columns

    # Called each time a commit has been analyzed,
    # and it is time to present the data in the commit.
    def add_row(self, columns, date, column_to_lines_dict):
        self.rows.append((date, column_to_lines_dict))

    # Called when the processing of commits is complete.
    # This is a good time to write output files to disk.
    def finish(self):
        # Import right before needed to significantly
        # improve startup time of this tool
        import os
        import matplotlib.pyplot as plt
        import numpy as np
        from .utils import print_file_written

        dates = list(map(lambda row: row[0], self.rows))

        line_counts = []
        for column in self.columns:
            line_count = list(
                map(lambda row: row[1].get(column, 0), self.rows))
            line_counts.append(line_count)

        s = np.vstack(line_counts)

        # See https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
        matplotlib_style = "dark_background"
        if self.args.style == "light":
            matplotlib_style = "default"

        # TODO: Validate arg earlier
        width_inches, height_inches = self.args.size_inches.split(':')
        width_inches = float(width_inches)
        height_inches = float(height_inches)

        plt.style.use(matplotlib_style)
        plt.figure(figsize=(width_inches, height_inches))
        plt.stackplot(dates, s, labels=self.columns)
        plt.legend(loc='upper left')
        if self.args.relative:
            plt.ylabel("Language usage %")
        else:
            plt.ylabel("Total (stacked) line count")
        plt.title(f"{os.path.basename(os.getcwd())} language trends")
        plt.tick_params(axis='x', labelrotation=45)
        plt.tight_layout()

        # TODO: Support stdout output
        plt.savefig(self.args.output)
        print_file_written(self.args.output)



import sys
import time

RATE_LIMIT_INTERVAL_SECONDS = 0.1


class Progress:

    def __init__(self, args, total_commits):
        self.args = args

        self.current_commit = 1
        self.total_commits = total_commits

        self.last_print = None

    def print_state(self, current_file, total_files):
        if (self.args.no_progress):
            return

        if (not sys.stderr.isatty):
            return

        # If we recently printed, bail out. Always print if this is the last file we
        # are processig however, since otherwise output seems "incomplete" to a human.
        if self.is_rate_limited() and current_file < total_files:
            return

        if self.total_commits == 1:
            commit_part = ""
        else:
            # "commit  12/345 "
            commit_part = f"commit {padded_progress(self.current_commit,self.total_commits)} "

        # "file  67/890"
        file_part = f"file {padded_progress(current_file,total_files)}"

        # "Counting lines in commit  12/345 file  67/890"
        print(
            f"Counting lines in {commit_part}{file_part}\r",
            file=sys.stderr,
            end='',
        )

    # Avoid writing large amounts of data to stderr, which can slow down execution significantly
    def is_rate_limited(self):
        now = time.time()
        if self.last_print is not None and now < self.last_print + RATE_LIMIT_INTERVAL_SECONDS:
            return True
        self.last_print = now
        return False

    def commit_processed(self):
        self.current_commit += 1


def padded_progress(current_commit, total_commits):
    pad = len(str(total_commits))
    return f"{current_commit:>{pad}}/{total_commits}"





def get_extensions_sorted_by_popularity(column_to_lines_dict):
    r = sorted(column_to_lines_dict, key=column_to_lines_dict.get)
    r.reverse()
    return r


# Excludes some extensions very unlikely to be of interest, e.g. '.lock'
def get_top_three_extensions(column_to_lines_dict):
    data = get_extensions_sorted_by_popularity(column_to_lines_dict)
    filtered_data = list(filter(lambda e: e != ".lock", data))
    return filtered_data[0:3]


def to_relative_numbers_if_enabled(args, column_to_lines):
    if not args.relative:
        return column_to_lines

    relative_column_to_lines = {}

    total_lines = float(sum(column_to_lines.values()))
    for column in list(column_to_lines.keys()):
        relative_column_to_lines[column] = round((column_to_lines[column] / total_lines) * 100, 1)

    return relative_column_to_lines


def print_file_written(filename):
    print(f"""
Wrote output to file:

    {filename}
""")
